/*
 * Copyright (c) 2014 Jan Kolarik
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @file ar9271.h
 *
 * Header file for AR9271 USB WiFi dongle.
 *
 */

#ifndef AR9271_H_
#define AR9271_H_

#include <usb/dev/driver.h>

#include "htc.h"

/** Number of transmission queues */
#define AR9271_QUEUES_COUNT 10

/** Number of GPIO pin used for handling led light */
#define AR9271_LED_PIN 15

/** AR9271 Registers */
typedef enum {
	/* ATH command register */
	AR9271_COMMAND = 0x0008,
	AR9271_COMMAND_RX_ENABLE = 0x00000004,
	
	/* ATH config register */
	AR9271_CONFIG = 0x0014,
	AR9271_CONFIG_ADHOC = 0x00000020,
	
	AR9271_QUEUE_BASE_MASK = 0x1000,
	
	/* EEPROM Addresses */
	AR9271_EEPROM_BASE = 0x2100,
	AR9271_EEPROM_MAC_ADDR_START = 0x2118,
	
	/* Reset MAC interface */
	AR9271_RC = 0x4000,
	AR9271_RC_AHB = 0x00000001,
		
	/* GPIO registers */
	AR9271_GPIO_IN_OUT = 0x4048,		/**< GPIO value read/set  */
	AR9271_GPIO_OE_OUT = 0x404C,		/**< GPIO set to output  */
	AR9271_GPIO_OE_OUT_ALWAYS = 0x3,	/**< GPIO always drive output */
	AR9271_GPIO_OUT_MUX1 = 0x4060,
	AR9271_GPIO_OUT_MUX2 = 0x4064,
	AR9271_GPIO_OUT_MUX3 = 0x4068,
	AR9271_GPIO_OUT_MUX_AS_OUT = 0x0,	/**< GPIO set mux as output */
    
	/* Wakeup related registers */
	AR9271_RTC_RC = 0x7000,
	AR9271_RTC_RC_MAC_WARM = 0x00000001,
	AR9271_RTC_RC_MAC_COLD = 0x00000002,
	AR9271_RTC_RC_MASK = 0x00000003,
	AR9271_RTC_PLL_CONTROL = 0x7014,
	AR9271_RTC_RESET = 0x7040,
	AR9271_RTC_STATUS = 0x7044,
	AR9271_RTC_STATUS_MASK = 0x0000000F,
	AR9271_RTC_STATUS_SHUTDOWN = 0x00000001,
	AR9271_RTC_STATUS_ON = 0x00000002,
	AR9271_RTC_FORCE_WAKE = 0x704C,
	AR9271_RTC_FORCE_WAKE_ENABLE = 0x00000001,
	AR9271_RTC_FORCE_WAKE_ON_INT = 0x00000002,
		
	/* MAC Registers */
	AR9271_STATION_ID0 = 0x8000,	/**< STA Address Lower 32 Bits */
	AR9271_STATION_ID1 = 0x8004,	/**< STA Address Upper 16 Bits */
	AR9271_BSSID0 = 0x8008,			/**< BSSID Lower 32 Bits */
	AR9271_BSSID1 = 0x800C,			/**< BSSID Upper 16 Bits */
	AR9271_BSSID_MASK0 = 0x80E0,		/**< BSSID Mask Lower 32 Bits */
	AR9271_BSSID_MASK1 = 0x80E4,		/**< BSSID Mask Upper 16 Bits */
	AR9271_STATION_ID1_MASK = 0x0000FFFF,
		
	/* RX filtering register */
	AR9271_RX_FILTER = 0x803C,
	AR9271_RX_FILTER_UNI = 0x00000001,
	AR9271_RX_FILTER_MULTI = 0x00000002,
	AR9271_RX_FILTER_BROAD = 0x00000004,
	AR9271_RX_FILTER_CONTROL = 0x00000008,
	AR9271_RX_FILTER_BEACON = 0x00000010,
	AR9271_RX_FILTER_PROMISCUOUS = 0x00000020,
	AR9271_RX_FILTER_PROBEREQ = 0x00000080,
	AR9271_RX_FILTER_MYBEACON = 0x00000200,
		
	/* Physical layer registers */
	AR9271_PHY_BASE = 0x9800,
	AR9271_PHY_ACTIVE = 0x981C,
	AR9271_ADC_CONTROL = 0x982C,
	AR9271_AGC_CONTROL = 0x9860,
	AR9271_PHY_MODE = 0xA200,
	AR9271_PHY_CCK_TX_CTRL = 0xA204,
	AR9271_PHY_TPCRG1 = 0xA258, 
	AR9271_CARRIER_LEAK_CONTROL = 0xA358,
	AR9271_ADC_CONTROL_OFF_PWDADC = 0x00008000,
	AR9271_AGC_CONTROL_CALIB = 0x00000001,
	AR9271_AGC_CONTROL_TX_CALIB = 0x00010000,
	AR9271_PHY_MODE_2G = 0x02,
	AR9271_PHY_MODE_DYNAMIC = 0x04,
	AR9271_PHY_CCK_TX_CTRL_JAPAN = 0x00000010,
	AR9271_PHY_TPCRG1_PD_CALIB = 0x00400000,
	AR9271_CARRIER_LEAK_CALIB = 0x00000002,
		
	AR9271_OPMODE_STATION_AP_MASK =	0x00010000,
	AR9271_OPMODE_ADHOC_MASK = 0x00020000,
		
	AR9271_CLOCK_CONTROL = 0x50040,
	AR9271_MAX_CPU_CLOCK = 0x304,
		
	AR9271_RESET_POWER_DOWN_CONTROL = 0x50044,
	AR9271_RADIO_RF_RESET = 0x20,
	AR9271_GATE_MAC_CONTROL = 0x4000,
    
	/* FW Addresses */
	AR9271_FW_ADDRESS = 0x501000,
	AR9271_FW_OFFSET = 0x903000,
} ar9271_registers_t;

/** AR9271 Requests */
typedef enum {
	AR9271_FW_DOWNLOAD = 0x30,
	AR9271_FW_DOWNLOAD_COMP = 0x31,
} ar9271_requests_t;

/** AR9271 device data */
typedef struct {
	/** Backing DDF device */
	ddf_dev_t *ddf_dev;
	
	/** USB device data */
	usb_device_t *usb_device;
	
	/** IEEE 802.11 device data */
	ieee80211_dev_t *ieee80211_dev;
	
	/** ATH device data */
	ath_t *ath_device;
	
	/** HTC device data */
	htc_device_t *htc_device;
} ar9271_t;

/**
 * AR9271 hardware init values.
 * 
 * Taken from Linux sources, some values omitted.
 */
static const uint32_t ar9271_init_array[][2] = {
	{0x0000000c, 0x00000000},
	{0x00000030, 0x00020045},
	{0x00000034, 0x00000005},
	{0x00000040, 0x00000000},
	{0x00000044, 0x00000008},
	{0x00000048, 0x00000008},
	{0x0000004c, 0x00000010},
	{0x00000050, 0x00000000},
	{0x00000054, 0x0000001f},
	{0x00000800, 0x00000000},
	{0x00000804, 0x00000000},
	{0x00000808, 0x00000000},
	{0x0000080c, 0x00000000},
	{0x00000810, 0x00000000},
	{0x00000814, 0x00000000},
	{0x00000818, 0x00000000},
	{0x0000081c, 0x00000000},
	{0x00000820, 0x00000000},
	{0x00000824, 0x00000000},
	{0x00001040, 0x002ffc0f},
	{0x00001044, 0x002ffc0f},
	{0x00001048, 0x002ffc0f},
	{0x0000104c, 0x002ffc0f},
	{0x00001050, 0x002ffc0f},
	{0x00001054, 0x002ffc0f},
	{0x00001058, 0x002ffc0f},
	{0x0000105c, 0x002ffc0f},
	{0x00001060, 0x002ffc0f},
	{0x00001064, 0x002ffc0f},
	{0x00001230, 0x00000000},
	{0x00001270, 0x00000000},
	{0x00001038, 0x00000000},
	{0x00001078, 0x00000000},
	{0x000010b8, 0x00000000},
	{0x000010f8, 0x00000000},
	{0x00001138, 0x00000000},
	{0x00001178, 0x00000000},
	{0x000011b8, 0x00000000},
	{0x000011f8, 0x00000000},
	{0x00001238, 0x00000000},
	{0x00001278, 0x00000000},
	{0x000012b8, 0x00000000},
	{0x000012f8, 0x00000000},
	{0x00001338, 0x00000000},
	{0x00001378, 0x00000000},
	{0x000013b8, 0x00000000},
	{0x000013f8, 0x00000000},
	{0x00001438, 0x00000000},
	{0x00001478, 0x00000000},
	{0x000014b8, 0x00000000},
	{0x000014f8, 0x00000000},
	{0x00001538, 0x00000000},
	{0x00001578, 0x00000000},
	{0x000015b8, 0x00000000},
	{0x000015f8, 0x00000000},
	{0x00001638, 0x00000000},
	{0x00001678, 0x00000000},
	{0x000016b8, 0x00000000},
	{0x000016f8, 0x00000000},
	{0x00001738, 0x00000000},
	{0x00001778, 0x00000000},
	{0x000017b8, 0x00000000},
	{0x000017f8, 0x00000000},
	{0x0000103c, 0x00000000},
	{0x0000107c, 0x00000000},
	{0x000010bc, 0x00000000},
	{0x000010fc, 0x00000000},
	{0x0000113c, 0x00000000},
	{0x0000117c, 0x00000000},
	{0x000011bc, 0x00000000},
	{0x000011fc, 0x00000000},
	{0x0000123c, 0x00000000},
	{0x0000127c, 0x00000000},
	{0x000012bc, 0x00000000},
	{0x000012fc, 0x00000000},
	{0x0000133c, 0x00000000},
	{0x0000137c, 0x00000000},
	{0x000013bc, 0x00000000},
	{0x000013fc, 0x00000000},
	{0x0000143c, 0x00000000},
	{0x0000147c, 0x00000000},
	{0x00004030, 0x00000002},
	{0x0000403c, 0x00000002},
	{0x00004024, 0x0000001f},
	{0x00004060, 0x00000000},
	{0x00004064, 0x00000000},
	{0x00008018, 0x00000700},
	{0x00008020, 0x00000000},
	{0x00008038, 0x00000000},
	{0x00008048, 0x00000000},
	{0x00008054, 0x00000000},
	{0x00008058, 0x00000000},
	{0x0000805c, 0x000fc78f},
	{0x00008060, 0x0000000f},
	{0x00008064, 0x00000000},
	{0x00008070, 0x00000000},
	{0x000080b0, 0x00000000},
	{0x000080b4, 0x00000000},
	{0x000080b8, 0x00000000},
	{0x000080bc, 0x00000000},
	{0x000080c0, 0x2a80001a},
	{0x000080c4, 0x05dc01e0},
	{0x000080c8, 0x1f402710},
	{0x000080cc, 0x01f40000},
	{0x000080d0, 0x00001e00},
	{0x000080d4, 0x00000000},
	{0x000080d8, 0x00400000},
	{0x000080e0, 0xffffffff},
	{0x000080e4, 0x0000ffff},
	{0x000080e8, 0x003f3f3f},
	{0x000080ec, 0x00000000},
	{0x000080f0, 0x00000000},
	{0x000080f4, 0x00000000},
	{0x000080f8, 0x00000000},
	{0x000080fc, 0x00020000},
	{0x00008100, 0x00020000},
	{0x00008104, 0x00000001},
	{0x00008108, 0x00000052},
	{0x0000810c, 0x00000000},
	{0x00008110, 0x00000168},
	{0x00008118, 0x000100aa},
	{0x0000811c, 0x00003210},
	{0x00008120, 0x08f04810},
	{0x00008124, 0x00000000},
	{0x00008128, 0x00000000},
	{0x0000812c, 0x00000000},
	{0x00008130, 0x00000000},
	{0x00008134, 0x00000000},
	{0x00008138, 0x00000000},
	{0x0000813c, 0x00000000},
	{0x00008144, 0xffffffff},
	{0x00008168, 0x00000000},
	{0x0000816c, 0x00000000},
	{0x00008170, 0x32143320},
	{0x00008174, 0xfaa4fa50},
	{0x00008178, 0x00000100},
	{0x0000817c, 0x00000000},
	{0x000081c0, 0x00000000},
	{0x000081d0, 0x0000320a},
	{0x000081ec, 0x00000000},
	{0x000081f0, 0x00000000},
	{0x000081f4, 0x00000000},
	{0x000081f8, 0x00000000},
	{0x000081fc, 0x00000000},
	{0x00008200, 0x00000000},
	{0x00008204, 0x00000000},
	{0x00008208, 0x00000000},
	{0x0000820c, 0x00000000},
	{0x00008210, 0x00000000},
	{0x00008214, 0x00000000},
	{0x00008218, 0x00000000},
	{0x0000821c, 0x00000000},
	{0x00008220, 0x00000000},
	{0x00008224, 0x00000000},
	{0x00008228, 0x00000000},
	{0x0000822c, 0x00000000},
	{0x00008230, 0x00000000},
	{0x00008234, 0x00000000},
	{0x00008238, 0x00000000},
	{0x0000823c, 0x00000000},
	{0x00008240, 0x00100000},
	{0x00008244, 0x0010f400},
	{0x00008248, 0x00000100},
	{0x0000824c, 0x0001e800},
	{0x00008250, 0x00000000},
	{0x00008254, 0x00000000},
	{0x00008258, 0x00000000},
	{0x0000825c, 0x400000ff},
	{0x00008260, 0x00080922},
	{0x00008264, 0x88a00010},
	{0x00008270, 0x00000000},
	{0x00008274, 0x40000000},
	{0x00008278, 0x003e4180},
	{0x0000827c, 0x00000000},
	{0x00008284, 0x0000002c},
	{0x00008288, 0x0000002c},
	{0x0000828c, 0x00000000},
	{0x00008294, 0x00000000},
	{0x00008298, 0x00000000},
	{0x0000829c, 0x00000000},
	{0x00008300, 0x00000040},
	{0x00008314, 0x00000000},
	{0x00008328, 0x00000000},
	{0x0000832c, 0x00000001},
	{0x00008330, 0x00000302},
	{0x00008334, 0x00000e00},
	{0x00008338, 0x00ff0000},
	{0x0000833c, 0x00000000},
	{0x00008340, 0x00010380},
	{0x00008344, 0x00581043},
	{0x00007010, 0x00000030},
	{0x00009808, 0x00000000},
	{0x0000980c, 0xafe68e30},
	{0x00009810, 0xfd14e000},
	{0x00009814, 0x9c0a9f6b},
	{0x0000981c, 0x00000000},
	{0x0000982c, 0x0000a000},
	{0x00009830, 0x00000000},
	{0x0000983c, 0x00200400},
	{0x0000984c, 0x0040233c},
	{0x00009854, 0x00000044},
	{0x00009900, 0x00000000},
	{0x00009904, 0x00000000},
	{0x00009908, 0x00000000},
	{0x0000990c, 0x00000000},
	{0x0000991c, 0x10000fff},
	{0x00009920, 0x04900000},
	{0x00009928, 0x00000001},
	{0x0000992c, 0x00000004},
	{0x00009934, 0x1e1f2022},
	{0x00009938, 0x0a0b0c0d},
	{0x0000993c, 0x00000000},
	{0x00009940, 0x14750604},
	{0x00009948, 0x9280c00a},
	{0x0000994c, 0x00020028},
	{0x00009954, 0x5f3ca3de},
	{0x00009958, 0x0108ecff},
	{0x00009968, 0x000003ce},
	{0x00009970, 0x192bb514},
	{0x00009974, 0x00000000},
	{0x00009978, 0x00000001},
	{0x0000997c, 0x00000000},
	{0x00009980, 0x00000000},
	{0x00009984, 0x00000000},
	{0x00009988, 0x00000000},
	{0x0000998c, 0x00000000},
	{0x00009990, 0x00000000},
	{0x00009994, 0x00000000},
	{0x00009998, 0x00000000},
	{0x0000999c, 0x00000000},
	{0x000099a0, 0x00000000},
	{0x000099a4, 0x00000001},
	{0x000099a8, 0x201fff00},
	{0x000099ac, 0x2def0400},
	{0x000099b0, 0x03051000},
	{0x000099b4, 0x00000820},
	{0x000099dc, 0x00000000},
	{0x000099e0, 0x00000000},
	{0x000099e4, 0xaaaaaaaa},
	{0x000099e8, 0x3c466478},
	{0x000099ec, 0x0cc80caa},
	{0x000099f0, 0x00000000},
	{0x0000a208, 0x803e68c8},
	{0x0000a210, 0x4080a333},
	{0x0000a214, 0x00206c10},
	{0x0000a218, 0x009c4060},
	{0x0000a220, 0x01834061},
	{0x0000a224, 0x00000400},
	{0x0000a228, 0x000003b5},
	{0x0000a22c, 0x00000000},
	{0x0000a234, 0x20202020},
	{0x0000a238, 0x20202020},
	{0x0000a244, 0x00000000},
	{0x0000a248, 0xfffffffc},
	{0x0000a24c, 0x00000000},
	{0x0000a254, 0x00000000},
	{0x0000a258, 0x0ccb5380},
	{0x0000a25c, 0x15151501},
	{0x0000a260, 0xdfa90f01},
	{0x0000a268, 0x00000000},
	{0x0000a26c, 0x0ebae9e6},
	{0x0000a388, 0x0c000000},
	{0x0000a38c, 0x20202020},
	{0x0000a390, 0x20202020},
	{0x0000a39c, 0x00000001},
	{0x0000a3a0, 0x00000000},
	{0x0000a3a4, 0x00000000},
	{0x0000a3a8, 0x00000000},
	{0x0000a3ac, 0x00000000},
	{0x0000a3b0, 0x00000000},
	{0x0000a3b4, 0x00000000},
	{0x0000a3b8, 0x00000000},
	{0x0000a3bc, 0x00000000},
	{0x0000a3c0, 0x00000000},
	{0x0000a3c4, 0x00000000},
	{0x0000a3cc, 0x20202020},
	{0x0000a3d0, 0x20202020},
	{0x0000a3d4, 0x20202020},
	{0x0000a3e4, 0x00000000},
	{0x0000a3e8, 0x18c43433},
	{0x0000a3ec, 0x00f70081},
	{0x0000a3f0, 0x01036a2f},
	{0x0000a3f4, 0x00000000},
	{0x0000d270, 0x0d820820},
	{0x0000d35c, 0x07ffffef},
	{0x0000d360, 0x0fffffe7},
	{0x0000d364, 0x17ffffe5},
	{0x0000d368, 0x1fffffe4},
	{0x0000d36c, 0x37ffffe3},
	{0x0000d370, 0x3fffffe3},
	{0x0000d374, 0x57ffffe3},
	{0x0000d378, 0x5fffffe2},
	{0x0000d37c, 0x7fffffe2},
	{0x0000d380, 0x7f3c7bba},
	{0x0000d384, 0xf3307ff0}
};

#endif